import sys

import vtk
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from PySide2.QtOpenGL import QGLWidget

from vmi import vio


class Mouse:
    def __init__(self):
        self._Enter = self._Leave = self._Wheel = self._Move = self.none
        self._LeftButtonPress = self._LeftButtonPressMove = self.none
        self._LeftButtonPressMoveRelease = self._LeftButtonPressRelease = self.none
        self._MidButtonPress = self._MidButtonPressMove = self.none
        self._MidButtonPressMoveRelease = self._MidButtonPressRelease = self.none
        self._RightButtonPress = self._RightButtonPressMove = self.none
        self._RightButtonPressMoveRelease = self._RightButtonPressRelease = self.none

        self.mouse = {
            'NoButton': {
                'Enter': self._Enter,
                'Leave': self._Leave,
                'Wheel': self._Wheel,
                'Move': self._Move},
            'LeftButton': {
                'Press': self._LeftButtonPress,
                'PressMove': self._LeftButtonPressMove,
                'PressMoveRelease': self._LeftButtonPressMoveRelease,
                'PressRelease': self._LeftButtonPressRelease},
            'MidButton': {
                'Press': self._MidButtonPress,
                'PressMove': self._MidButtonPressMove,
                'PressMoveRelease': self._MidButtonPressMoveRelease,
                'PressRelease': self._MidButtonPressRelease},
            'RightButton': {
                'Press': self._RightButtonPress,
                'PressMove': self._RightButtonPressMove,
                'PressMoveRelease': self._RightButtonPressMoveRelease,
                'PressRelease': self._RightButtonPressRelease}}

    @staticmethod
    def none(**kwargs): pass


class Widget(Mouse, QGLWidget):
    """vtk场景"""

    def __init__(self, parent=None):
        Mouse.__init__(self)
        QGLWidget.__init__(self, parent)

        rect = QGuiApplication.screens()[0].geometry()
        rect = QRect(rect.center() - QPoint(800, 450), QSize(1600, 900))
        self.setGeometry(rect)
        self.setAttribute(Qt.WA_OpaquePaintEvent)
        self.setAttribute(Qt.WA_PaintOnScreen)
        self.setFocusPolicy(Qt.WheelFocus)
        self.setMouseTracking(True)

        # 渲染器
        self._Renderer = vtk.vtkRenderer()
        self._Renderer.SetUseHiddenLineRemoval(1)  # 隐线消除
        self._Renderer.SetUseFXAA(False)  # 快速近似抗锯齿

        # 渲染窗口
        self._RenderWindow = vtk.vtkRenderWindow()
        self._RenderWindow.SetWindowInfo(str(int(self.winId())))
        self._RenderWindow.AddRenderer(self._Renderer)

        # 交互器
        self._Interactor = vtk.vtkGenericRenderWindowInteractor()
        self._Interactor.SetRenderWindow(self._RenderWindow)
        self._Interactor.Initialize()

        # 方位标
        self._OrientationMarker = init_orientation_marker(self._Interactor)

        # 照相机
        self._Camera = self.camera()
        self.camera().SetParallelProjection(True)  # 平行投影

        # 拾取器
        self._Picker = vtk.vtkPicker()
        self._PickerProp = vtk.vtkPropPicker()
        self._PickerPoint = vtk.vtkPointPicker()
        self._PickerCell = vtk.vtkCellPicker()
        self._Picker.SetTolerance(0.005)
        self._PickerPoint.SetTolerance(0.005)
        self._PickerCell.SetTolerance(0.005)

        # 刷新计时器
        self._Timer = QTimer(self)
        self._Timer.timeout.connect(self.timeout)
        self._Timer.start(10)

        self._Time = QTime()
        self._Time.start()
        self._FPS = self._UpdateSum = 0

        # 创建一个隐藏控件，当销毁时调用清理vtk元素
        self._Hidden = QWidget(self)
        self._Hidden.hide()
        self._Hidden.destroyed.connect(self._Finalize)

        # 更新请求
        self._UpdateAtOnce: bool = False

        # 鼠标位置
        self._MousePosLast: QPoint = QPoint(0, 0)
        self._MousePos: QPoint = QPoint(0, 0)

        # 鼠标正在交互的对象
        self._PropEnter: vtk.vtkProp = None
        self._ViEnter: Widget = None

        # 鼠标按键状态
        self._ButtonPress = self._ButtonPressMove = 'NoButton'

        # 右键菜单
        self.setMouseStyle('volume')
        self.menu = QMenu(self.tr('Window'))
        self.menu.addAction(self.tr('Fit content')).triggered.connect(self.resetCamera)

    def menuExec(self, **kwargs):
        self.menu.exec_(QCursor.pos())

    def __setstate__(self, s):
        self.__init__()
        self.__dict__.update(s)

        for kw in s['_VTKGets']:
            vio.vtkset(getattr(self, kw), s['_VTKGets'][kw])

    def __getstate__(self):
        s = self.__dict__.copy()

        s['_VTKGets'] = {name: vio.vtkget(getattr(self, name)) for name in
                         ['_RenderWindow', '_Renderer', '_Interactor', '_Camera',
                          '_Picker', '_PickerProp', '_PickerPoint', '_PickerCell']}

        for kw in s['_VTKGets']:
            del s[kw]
        del s['_OrientationMarker'], s['_Timer'], s['_Hidden'], s['__METAOBJECT__']
        return s

    def _Finalize(self):
        self._Timer.stop()
        self._RenderWindow.Finalize()

    def camera(self, *args):
        if len(args) == 0:
            return self._Renderer.GetActiveCamera()
        elif len(args) > 0:
            fp = list(self.camera().GetFocalPoint())
            cp = list(self.camera().GetPosition())
            up = list(self.camera().GetViewUp())

            look, right = [fp[i] - cp[i] for i in range(3)], [0] * 3

            vtk.vtkMath.Cross(look, up, right)
            vtk.vtkMath.Normalize(look)
            vtk.vtkMath.Normalize(right)
            vtk.vtkMath.Normalize(up)

            t, value = [], {'fp': fp, 'cp': cp, 'up': up, 'look': look, 'right': right}
            for key in args:
                if key in value.keys():
                    t.append(value[key])
            return t

    def closeEvent(self, ev):
        self._Finalize()

    def entered(self):
        """返回鼠标悬停状态，在物体上->prop/在场景内->vmi/在场景外->None"""
        return self._PropEnter if self._PropEnter else (self._ViEnter if self._ViEnter else self)

    def enterEvent(self, ev):
        self._ViEnter = self
        c, b, e = self, 'NoButton', 'Enter'
        c.mouse[b][e](caller=c, event=e, button=b)

        self._PropEnter, *_ = self.pickProp(ev.pos())
        if self._PropEnter:
            c = self._PropEnter
            c.mouse[b][e](caller=c, event=e, button=b)

    def leaveEvent(self, ev):
        if self._ViEnter:
            c, b, e = self, 'NoButton', 'Leave'
            c.mouse[b][e](caller=c, event=e, button=b)
        if self._PropEnter:
            c, b, e = self._PropEnter, 'NoButton', 'Leave'
            c.mouse[b][e](caller=c, event=e, button=b)
        self._ViEnter = self._PropEnter = None

    def menuWindowFitContent(self):
        print(self)

    def mouseButtonRotateFocal(self, **kwargs):
        if kwargs['event'] == 'PressMove':
            dx, dy = self.mouseOverDisplay()
            rx, ry = -200 * dx / self.height(), 200 * dy / self.width()
            self.camera().Azimuth(rx)
            self.camera().Elevation(ry)
            self.camera().OrthogonalizeViewUp()
            self.updateInTime()

    def mouseButtonPan(self, **kwargs):
        if kwargs['event'] == 'PressMove':
            dp = self.mouseOverFPlane()
            fp, cp = self.camera('fp', 'cp')
            fp = [fp[i] - dp[i] for i in range(3)]
            cp = [cp[i] - dp[i] for i in range(3)]
            self.camera().SetFocalPoint(fp)
            self.camera().SetPosition(cp)

            if self._Interactor.GetLightFollowCamera():
                self._Renderer.UpdateLightsGeometryToFollowCamera()

            self.updateInTime()

    def mouseButtonZoom(self, **kwargs):
        if kwargs['event'] == 'PressMove':
            c = self._Renderer.GetCenter()
            _, dy = self.mouseOverDisplay()
            factor = 1.1 ** (10 * -dy / c[1])

            if self.camera().GetParallelProjection():
                self.camera().SetParallelScale(self.camera().GetParallelScale() / factor)
            else:
                self.camera().Dolly(factor)

            if self._Interactor.GetLightFollowCamera():
                self._Renderer.UpdateLightsGeometryToFollowCamera()

            self.updateInTime()

    def mouseMoveEvent(self, ev):
        self._MousePosLast, self._MousePos = self._MousePos, ev.pos()

        if self._ButtonPress == 'NoButton':
            prop, *_ = self.pickProp(ev.pos())
            last_prop, self._PropEnter = self._PropEnter, prop

            if not last_prop and not prop:
                c, b, e = self, 'NoButton', 'Move'
                c.mouse[b][e](caller=c, event=e, button=b)
            elif not last_prop and prop:
                c, b, e = prop, 'NoButton', 'Enter'
                c.mouse[b][e](caller=c, event=e, button=b)
            elif last_prop and not prop:
                c, b, e = last_prop, 'NoButton', 'Leave'
                c.mouse[b][e](caller=c, event=e, button=b)
            elif last_prop and prop and last_prop is prop:
                c, b, e = prop, 'NoButton', 'Move'
                c.mouse[b][e](caller=c, event=e, button=b)
            elif last_prop and prop and last_prop is not prop:
                c, b, e = last_prop, 'NoButton', 'Leave'
                c.mouse[b][e](caller=c, event=e, button=b)
                c, b, e = prop, 'NoButton', 'Enter'
                c.mouse[b][e](caller=c, event=e, button=b)
        elif self._MousePosLast != self._MousePos:
            self._ButtonPressMove = self._ButtonPress
            c, b, e = self.entered(), self._ButtonPressMove, 'PressMove'
            c.mouse[b][e](caller=c, event=e, button=b)

    def mouseOverDisplay(self):
        """返回鼠标位移向量
        -> (dx, dy)"""
        return self._MousePos.x() - self._MousePosLast.x(), self._MousePos.y() - self._MousePosLast.y()

    def mouseOverFPlane(self):
        """返回鼠标在焦平面上的位移向量
        -> (x, y, z)"""
        pt = self.pickFPlane(self._MousePos)
        ptLast = self.pickFPlane(self._MousePosLast)
        return [pt[i] - ptLast[i] for i in range(3)]

    def mousePos(self):
        """返回当前鼠标位置"""
        return self._MousePos

    def mousePosLast(self):
        """返回上一鼠标位置"""
        return self._MousePosLast

    def mousePressEvent(self, ev):
        self._MousePosLast, self._MousePos = self._MousePos, ev.pos()

        if self._ButtonPress == 'NoButton':
            self._ButtonPress = ev.button().name.decode()
            d = ev.type() == QEvent.Type.MouseButtonDblClick
            c, b, e = self.entered(), self._ButtonPress, 'Press'
            c.mouse[b][e](caller=c, event=e, button=b, double=d)

    def mouseReleaseEvent(self, ev):
        self._MousePosLast, self._MousePos = self._MousePos, ev.pos()

        if ev.button().name.decode() == self._ButtonPressMove:
            c, b, e = self.entered(), self._ButtonPressMove, 'PressMoveRelease'
            c.mouse[b][e](caller=c, event=e, button=b)
            self._ButtonPress = self._ButtonPressMove = 'NoButton'
        elif ev.button().name.decode() == self._ButtonPress:
            c, b, e = self.entered(), self._ButtonPress, 'PressRelease'
            c.mouse[b][e](caller=c, event=e, button=b)
            self._ButtonPress = self._ButtonPressMove = 'NoButton'

    def paintEngine(self):
        return None

    def paintEvent(self, ev):
        self._Interactor.Render()

    def pick(self, pos: QPoint, picker):
        """拾取场景对象"""

        picker.Pick(pos.x(), self.height() - pos.y(), 0, self._Renderer)
        prop = picker.GetViewProp()
        prop = prop._Prop if prop else None
        return prop, picker.GetPickPosition(), picker

    def pickBBox(self, pos: QPoint):
        """拾取场景对象的包围盒
        pickBBox(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._Picker)

    def pickCell(self, pos: QPoint):
        """拾取场景对象的单元
        pickCell(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._PickerCell)

    def pickFPlane(self, pos: QPoint):
        """拾取场景的焦平面
        pickFPlane(QPoint(x, y)) -> (x, y, z)"""
        t = [0] * 4
        vtk.vtkInteractorObserver.ComputeDisplayToWorld(
            self._Renderer, pos.x(), self.height() - pos.y(), 0, t)
        pt = [_ / t[3] for _ in t[:3]]

        fp, look = self.camera('fp', 'look')
        vtk.vtkPlane.ProjectPoint(pt, fp, look, pt)
        return pt

    def pickPoint(self, pos: QPoint):
        """拾取场景对象的点
        pickPoint(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._PickerPoint)

    def pickProp(self, pos: QPoint):
        """拾取场景对象
        pickProp(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._PickerProp)

    def pickToDisplay(self, pt: list):
        """拾取的场景坐标转换为屏幕坐标
        pickToDisplay((x, y, z) -> QPoint(x, y)"""
        t = [0] * 3
        vtk.vtkInteractorObserver.ComputeWorldToDisplay(
            self._Renderer, pt[0], pt[1], pt[2], t)
        return QPoint(t[0], self.height() - t[1])

    def resetCamera(self, b=None):
        if b is not None:
            self._Renderer.ResetCamera()
        else:
            self._Renderer.ResetCamera(b)
        self.updateInTime()

    def resizeEvent(self, ev):
        w = self.width()
        h = self.height()
        vtk.vtkRenderWindow.SetSize(self._RenderWindow, w, h)
        self._Interactor.SetSize(w, h)
        self._Interactor.ConfigureEvent()
        self.update()

    def setMouseStyle(self, style):
        if style.lower() == 'none':
            for b in self.mouse:
                for e in self.mouse[b]:
                    self.mouse[b][e] = self.none
        elif style.lower() == 'volume':
            self.setMouseStyle('none')
            self.mouse['LeftButton']['PressMove'] = self.mouseButtonRotateFocal
            self.mouse['MidButton']['PressMove'] = self.mouseButtonPan
            self.mouse['RightButton']['PressMove'] = self.mouseButtonZoom
            self.mouse['RightButton']['PressRelease'] = self.menuExec
        elif style.lower() == 'slice':
            self.setMouseStyle('none')

    def setOrientationMarkerVisible(self, arg: bool):
        """显示/隐藏方位标"""
        self._OrientationMarker.GetOrientationMarker().SetVisibility(arg)
        self.updateInTime()

    def sizeHint(self):
        return QSize(1600, 900)

    def timeout(self):
        if self._UpdateAtOnce:
            self._UpdateAtOnce = False
        self.updateAtOnce()

        if self._Time.elapsed() > 0.5e3:
            self._FPS = 1000 * self._UpdateSum / self._Time.elapsed()
            self.setWindowTitle('FPS: {:.0f}'.format(self._FPS))
            self._Time.restart()
            self._UpdateSum = 0

    def updateInTime(self):
        """及时刷新视图"""
        self._UpdateAtOnce = True

    def updateAtOnce(self):
        """立即刷新视图"""
        self._Renderer.ResetCameraClippingRange()
        self._RenderWindow.Render()
        self._UpdateSum += 1

    def wheelEvent(self, ev):
        d = ev.delta() / 120
        c, b, e = self.entered(), 'NoButton', 'Wheel'
        c.mouse[b][e](caller=c, event=e, button=b, delta=d)


def init_orientation_marker(i):
    """
    :param i: vtk.vtkRenderWindowInteractor
    :return: vtk.vtkOrientationMarkerWidget
    """

    vtk.vtkMapper.SetResolveCoincidentTopologyToPolygonOffset()

    cube = vtk.vtkAnnotatedCubeActor()
    cube.SetFaceTextScale(0.65)
    cube.SetXPlusFaceText("L")
    cube.SetXMinusFaceText("R")
    cube.SetYPlusFaceText("P")
    cube.SetYMinusFaceText("A")
    cube.SetZPlusFaceText("S")
    cube.SetZMinusFaceText("I")

    prop = cube.GetCubeProperty()
    prop.SetColor(0.75, 0.75, 1)

    prop = cube.GetTextEdgesProperty()
    prop.SetLineWidth(1)
    prop.SetDiffuse(0)
    prop.SetAmbient(1)
    prop.SetColor(0, 0, 0)

    prop = cube.GetXPlusFaceProperty()
    prop.SetColor(1, 0, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetXMinusFaceProperty()
    prop.SetColor(1, 0, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetYPlusFaceProperty()
    prop.SetColor(0, 1, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetYMinusFaceProperty()
    prop.SetColor(0, 1, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetZPlusFaceProperty()
    prop.SetColor(0, 0, 1)
    prop.SetInterpolationToFlat()
    prop = cube.GetZMinusFaceProperty()
    prop.SetColor(0, 0, 1)
    prop.SetInterpolationToFlat()

    axes = vtk.vtkAxesActor()

    axes.SetShaftTypeToCylinder()
    axes.SetXAxisLabelText("X")
    axes.SetYAxisLabelText("Y")
    axes.SetZAxisLabelText("Z")

    axes.SetTotalLength(1.5, 1.5, 1.5)
    axes.SetCylinderRadius(0.500 * axes.GetCylinderRadius())
    axes.SetConeRadius(1.025 * axes.GetConeRadius())
    axes.SetSphereRadius(1.500 * axes.GetSphereRadius())

    prop = axes.GetXAxisCaptionActor2D().GetCaptionTextProperty()
    prop.BoldOn()
    prop.ItalicOn()
    prop.ShadowOn()
    prop.SetFontFamilyToArial()
    prop.SetColor(1, 0, 0)

    prop = axes.GetYAxisCaptionActor2D().GetCaptionTextProperty()
    prop.BoldOn()
    prop.ItalicOn()
    prop.ShadowOn()
    prop.SetFontFamilyToArial()
    prop.SetColor(0, 1, 0)

    prop = axes.GetZAxisCaptionActor2D().GetCaptionTextProperty()
    prop.BoldOn()
    prop.ItalicOn()
    prop.ShadowOn()
    prop.SetFontFamilyToArial()
    prop.SetColor(0, 0, 1)

    assembly = vtk.vtkPropAssembly()
    assembly.AddPart(axes)
    assembly.AddPart(cube)

    ort = vtk.vtkOrientationMarkerWidget()
    ort.SetOrientationMarker(assembly)
    ort.SetViewport(0.0, 0.0, 0.2, 0.2)
    ort.SetInteractor(i)
    ort.SetEnabled(1)
    ort.SetInteractive(0)
    ort.On()

    return ort


if __name__ == '__main__':
    app = QApplication(sys.argv)
    app.setOrganizationName('Vi')
    app.setApplicationName('Vi')

    w = Widget()
    w.show()
    # w.setMouseStyle('Volume')
    w._Renderer.SetBackground(1, 0.5, 0.5)

    # import shelve
    #
    # with shelve.open('C:/Users/Medraw/Desktop/d') as db:
    #     db['w'] = w
    #
    # with shelve.open('C:/Users/Medraw/Desktop/d') as db:
    #     v = db['w']
    #     v.show()

    sys.exit(app.exec_())



from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from vtk import *

from .vask import *
from .vdcm import *
from .vidget import *
from .vio import *
from .vprop import *

name = 'vmi'
version = '0.0.10'



import sys

import numpy.linalg as npl
import vtk

from vmi import vio
from vmi.vidget import Mouse, Widget


class _Prop(Mouse):
    """场景物体
    self.prop is a vtk.vtkProp"""

    def __init__(self, vi: Widget):
        Mouse.__init__(self)

        self._Vi = vi
        self.setMouseStyle('Vi')
        self._Prop = None

    def setMouseStyle(self, style):
        if style == 'None':
            for b in self.mouse:
                for e in self.mouse[b]:
                    self.mouse[b][e] = self.none
        elif style == 'Vi':
            for b in self.mouse:
                for e in self.mouse[b]:
                    self.mouse[b][e] = self._Vi.mouse[b][e]

    def pickable(self, arg=None):
        if arg is None:
            return self._Prop.GetPickable()
        else:
            self._Prop.SetPickable(arg)
            self._Vi.updateInTime()

    def visible(self, arg=None):
        if arg is None:
            return self._Prop.GetVisibility()
        else:
            self._Prop.SetVisibility(arg)
            self._Vi.updateInTime()


class _Dataset:
    """场景物体的数据表达
    self.data is vtk.vtkDataSet"""

    def __init__(self):
        self._Dataset = None

    def clone(self, other=None):
        if other is not None:
            self._Dataset.ShallowCopy(self._Dataset.__class__())
        if hasattr(other, '_Dataset'):
            self._Dataset.ShallowCopy(other._Dataset)
        if isinstance(other, self._Dataset.__class__):
            self._Dataset.ShallowCopy(other)
        else:
            raise TypeError(type(other))


class ImageSlice(_Prop, _Dataset):
    """场景表示，图像数据的断层表示
    vtk.vtkImageData -> vtk.vtkImageSlice"""

    def __init__(self, vi: Widget):
        _Prop.__init__(self, vi)
        _Dataset.__init__(self)

        self._Mapper = vtk.vtkImageResliceMapper()
        self._Prop = vtk.vtkImageSlice()
        self._Prop._Prop = self
        self._Vi._Renderer.AddViewProp(self._Prop)

        self._Property = self._Prop.GetProperty()
        self._Property.SetInterpolationTypeToCubic()
        self._Property.SetUseLookupTableScalarRange(1)

        self._Dataset = vtk.vtkImageData()
        self._BindDataset = self

        self._SlicePlane = vtk.vtkPlane()
        self._BindSlicePlane = self

        self._LookupTable = vtk.vtkLookupTable()
        self._BindLookupTable = self

        self._WindowWidth = self._WindowLevel = 400

        self.bindDataset()
        self.bindSlicePlane()
        self.bindLookupTable()

        self.mouse['NoButton']['Wheel'] = self.slice
        self.mouse['LeftButton']['PressMove'] = self.window

    def delete(self):
        self._Dataset.ReleaseData()
        self._Vi._Renderer.RemoveViewProp(self._Prop)
        self._Dataset = self._Mapper = self._Prop = self._Property = self._Vi = None
        self._BindDataset = self._BindSlicePlane = self._BindLookupTable

    def __setstate__(self, s):
        self.__init__(s['_Vi'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Dataset.DeepCopy(vio.loads(s['_Dumps']['_Dataset']))
        self._LookupTable.DeepCopy(vio.loads(s['_Dumps']['_LookupTable']))
        del s['_Dumps']

        for kw in s['_VTKGets']:
            vio.vtkset(getattr(self, kw), s['_VTKGets'][kw])
        del s['_VTKGets']

        self.bindDataset()
        self.bindSlicePlane()
        self.bindLookupTable()

    def __getstate__(self):
        s = self.__dict__.copy()

        s['_Dumps'] = {'_Dataset': vio.dumps(self._Dataset),
                       '_LookupTable': vio.dumps(self._LookupTable)}
        for kw in s['_Dumps']:
            del s[kw]

        s['_VTKGets'] = {name: vio.vtkget(getattr(self, name)) for name in
                         ['_Mapper', '_Prop', '_Property', '_SlicePlane']}
        for kw in s['_VTKGets']:
            del s[kw]

        return s

    def bindDataset(self, other=None):
        if hasattr(other, '_Dataset'):
            self._BindDataset = other
        self._Mapper.SetInputData(self._BindDataset._Dataset)
        self._Prop.SetMapper(self._Mapper)
        self._Vi.updateInTime()

    def bindSlicePlane(self, other=None):
        if hasattr(other, '_SlicePlane'):
            self._BindSlicePlane = other
        self._Mapper.SetSlicePlane(self._BindSlicePlane._SlicePlane)
        self._Vi.updateInTime()

    def bindLookupTable(self, other=None):
        if hasattr(other, '_LookupTable'):
            self._BindLookupTable = other
        self._Property.SetLookupTable(self._BindLookupTable._LookupTable)
        self._Vi.updateInTime()

    def setSlicePlane(self, origin=None, normal=None):
        if isinstance(origin, str):
            if origin.lower() == 'c':
                o = self._BindDataset._Dataset.GetCenter()
                self._BindSlicePlane._SlicePlane.SetOrigin(o[0], o[1], o[2])
            elif origin.lower() in ('b0', 'b1', 'b2', 'b3', 'b4', 'b5'):
                i = int(origin[-1])
                o = list(self._BindDataset._Dataset.GetCenter())
                o[int(i / 2)] = self._BindDataset._Dataset.GetBounds()[i]
                self._BindSlicePlane._SlicePlane.SetOrigin(o[0], o[1], o[2])
        elif origin is not None:
            self._BindSlicePlane._SlicePlane.SetOrigin(origin[0], origin[1], origin[2])
        if isinstance(normal, str):
            if normal.lower() == 'sagittal':
                self._BindSlicePlane._SlicePlane.SetNormal(1, 0, 0)
            elif normal.lower() == 'coronal':
                self._BindSlicePlane._SlicePlane.SetNormal(0, 1, 0)
            elif normal.lower() == 'axial':
                self._BindSlicePlane._SlicePlane.SetNormal(0, 0, 1)
        elif normal is not None:
            self._BindSlicePlane._SlicePlane.SetNormal(normal[0], normal[1], normal[2])
        self._Vi.updateInTime()

    def setWindow(self, preset=None, width=None, level=None):
        if preset is None:
            if width is not None:
                self._WindowWidth = width
            if level is not None:
                self._WindowLevel = level

            self._WindowWidth = int(self._WindowWidth)
            self._WindowLevel = int(self._WindowLevel)

            r = [self._WindowLevel - 0.5 * self._WindowWidth,
                 self._WindowLevel + 0.5 * self._WindowWidth]

            t = self._BindLookupTable._LookupTable
            t.SetNumberOfTableValues(self._WindowWidth)
            t.SetTableRange(r)
            t.SetBelowRangeColor(0, 0, 0, 1.0)
            t.SetAboveRangeColor(1, 1, 1, 1.0)
            t.SetUseBelowRangeColor(1)
            t.SetUseAboveRangeColor(1)

            for i in range(self._WindowWidth):
                v = i / self._WindowWidth
                t.SetTableValue(i, (v, v, v, 1.0))
            self._Vi.updateInTime()

        elif preset.lower() == 'auto':
            r = vtk.vtkImageHistogramStatistics()
            r.SetInputData(self._BindDataset._Dataset)
            r.SetAutoRangePercentiles(1, 99)
            r.SetGenerateHistogramImage(0)
            r.Update()
            r = r.GetAutoRange()
            self.setWindow(r[1] - r[0], 0.5 * (r[0] + r[1]))
        elif preset.lower() == 'bone':
            self.setWindow(width=1000, level=400)
        elif preset.lower() == 'soft':
            self.setWindow(width=350, level=50)

    def slice(self, **kwargs):
        o = list(self._BindSlicePlane._SlicePlane.GetOrigin())
        n = self._BindSlicePlane._SlicePlane.GetNormal()
        dim = self._BindDataset._Dataset.GetDimensions()
        dxyz = self._BindDataset._Dataset.GetSpacing()

        n = [_ / npl.norm(n) for _ in n]
        dn = n[0] * dxyz[0] + n[1] * dxyz[1] + n[2] * dxyz[2]

        for i in range(3):
            if dim[i] > 1:
                o[i] += kwargs['delta'] * dn * n[i]

        if vtk.vtkMath.PlaneIntersectsAABB(self._Dataset.GetBounds(), n, o) == 0:
            self.setSlicePlane(origin=o)

    def window(self, **kwargs):
        dx, dy = self._Vi.mouseOverDisplay()

        r = self._Dataset.GetScalarRange()
        t = (r[1] - r[0]) / 2048
        self._WindowWidth += t * dx
        self._WindowLevel -= t * dy

        self._WindowWidth = max(self._WindowWidth, 0)
        self._WindowWidth = min(self._WindowWidth, r[1] - r[0])
        self._WindowLevel = max(self._WindowLevel, r[0])
        self._WindowLevel = min(self._WindowLevel, r[1])
        self.setWindow()


class ImageVolume(_Prop, _Dataset):
    """场景表示，图像数据的立体表示
    vtk.vtkImageData -> vtk.vtkVolume"""

    def __init__(self, vi: Widget):
        _Prop.__init__(self, vi)
        _Dataset.__init__(self)

        self._Mapper = vtk.vtkGPUVolumeRayCastMapper()
        self._Prop = vtk.vtkVolume()
        self._Prop._Prop = self
        self._Vi._Renderer.AddVolume(self._Prop)

        self._Mapper.SetBlendModeToComposite()
        self._Mapper.SetMaxMemoryInBytes(4096)
        self._Mapper.SetMaxMemoryFraction(1)
        self._Mapper.SetAutoAdjustSampleDistances(0)
        self._Mapper.SetLockSampleDistanceToInputSpacing(1)
        self._Mapper.SetUseJittering(1)

        self._Property = self._Prop.GetProperty()
        self._Property.SetInterpolationTypeToLinear()
        self._Property.SetAmbient(0.1)
        self._Property.SetDiffuse(0.9)
        self._Property.SetShade(1)

        self._Dataset = vtk.vtkImageData()
        self._BindDataset = self

        self._Color = {0: (1, 1, 1)}
        self._ScalarOpacity = {0: 0, 400: 1}
        self._GradientOpacity = {}

        self.bindDataset()
        self.setColor()
        self.setScalarOpacity()
        self.setGradientOpacity()

    def delete(self):
        self._Dataset.ReleaseData()
        self._Vi._Renderer.RemoveVolume(self._Prop)

    def __setstate__(self, s):
        self.__init__(s['_Vi'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Dataset.DeepCopy(vio.loads(s['_Dumps']['_Dataset']))
        del s['_Dumps']

        for kw in s['_VTKGets']:
            vio.vtkset(getattr(self, kw), s['_VTKGets'][kw])
        del s['_VTKGets']

        self.bindDataset()
        self.setColor()
        self.setScalarOpacity()
        self.setGradientOpacity()

    def __getstate__(self):
        s = self.__dict__.copy()

        s['_Dumps'] = {'_Dataset': vio.dumps(self._Dataset),
                       '_Color': self._Color}
        for kw in s['_Dumps']:
            del s[kw]

        s['_VTKGets'] = {name: vio.vtkget(getattr(self, name)) for name in
                         ['_Mapper', '_Prop', '_Property']}
        for kw in s['_VTKGets']:
            del s[kw]

        return s

    def bindDataset(self, other=None):
        if hasattr(other, '_Dataset'):
            self._BindDataset = other
        self._Mapper.SetInputData(self._BindDataset._Dataset)
        self._Prop.SetMapper(self._Mapper)
        self._Vi.updateInTime()

    def setColor(self, scalar_rgb=None):
        if scalar_rgb is None:
            f = vtk.vtkColorTransferFunction()
            for x in self._Color:
                r, g, b = self._Color[x]
                f.AddRGBPoint(x, r, g, b)
            self._Property.SetColor(f)
        else:
            self._Color = scalar_rgb
            self.setColor()

    def setScalarOpacity(self, scalar_opacity=None):
        if scalar_opacity is None:
            f = vtk.vtkPiecewiseFunction()
            for x in self._ScalarOpacity:
                f.AddPoint(x, self._ScalarOpacity[x])
            self._Property.SetScalarOpacity(f)
        else:
            self._Color = scalar_opacity
            self.setScalarOpacity()

    def setGradientOpacity(self, gradient_opacity=None):
        if gradient_opacity is None:
            f = vtk.vtkPiecewiseFunction()
            for x in self._GradientOpacity:
                f.AddPoint(x, self._GradientOpacity[x])
            self._Property.SetGradientOpacity(f)
        else:
            self._Color = gradient_opacity
            self.setGradientOpacity()


class PolyActor(_Prop, _Dataset):
    """场景表示，聚合数据的表示
    vtk.vtkPolyData -> vtk.vtkActor"""

    def __init__(self, vi: Widget):
        _Prop.__init__(self, vi)
        _Dataset.__init__(self)

        self._Mapper = vtk.vtkPolyDataMapper()
        self._Prop = vtk.vtkActor()
        self._Prop._Prop = self
        self._Vi._Renderer.AddActor(self._Prop)

        self._Prop.SetPickable(1)  # 默认可拾取

        self._Property = self._Prop.GetProperty()
        self._Property.SetAmbient(0.1)
        self._Property.SetDiffuse(0.9)

        self._BackfaceProperty = self.propertyBackface()
        self._Prop.SetBackfaceProperty(self._Property)

        self._Dataset = vtk.vtkPolyData()
        self._BindDataset = self
        self.bindDataset()

    def delete(self):
        self._Dataset.ReleaseData()
        self._Vi._Renderer.RemoveActor(self._Prop)

    def bindDataset(self, other=None):
        if hasattr(other, '_Dataset'):
            self._BindDataset = other
        self._Mapper.SetInputData(self._BindDataset._Dataset)
        self._Prop.SetMapper(self._Mapper)
        self._Vi.updateInTime()

    def propertyBackface(self):
        return self._Prop.GetBackfaceProperty()

    def __setstate__(self, s):
        self.__init__(s['_Vi'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Dataset.DeepCopy(vio.loads(s['_Dumps']['_Dataset']))
        del s['_Dumps']

        for kw in s['_VTKGets']:
            vio.vtkset(getattr(self, kw), s['_VTKGets'][kw])
        del s['_VTKGets']

        self.bindDataset()

    def __getstate__(self):
        s = self.__dict__.copy()

        s['_Dumps'] = {'_Dataset': vio.dumps(self._Dataset)}
        for kw in s['_Dumps']:
            del s[kw]

        s['_VTKGets'] = {name: vio.vtkget(getattr(self, name)) for name in
                         ['_Mapper', '_Prop', '_Property', '_BackfaceProperty']}
        for kw in s['_VTKGets']:
            del s[kw]

        return s


class TextActor(_Prop):
    """场景表示，聚合数据的表示
        vtk.vtkPolyData -> vtk.vtkActor"""

    def __init__(self, vi: Widget):
        _Prop.__init__(self, vi)
        self.prop = vtk.vtkTextActor()


if __name__ == '__main__':
    from PySide2.QtWidgets import *
    from pprint import pprint

    app = QApplication(sys.argv)
    app.setOrganizationName('Vi')
    app.setApplicationName('Vi')

    w = Widget()
    w.setMouseStyle("Volume")
    w.show()

    reader = vtk.vtkSTLReader()
    reader.SetFileName('C:/Users/Medraw/Desktop/0921.stl')
    reader.Update()

    pa = PolyActor(w)
    pa.clone(reader.GetOutput())
    w._Renderer.ResetCamera()

    pa1 = PolyActor(w)
    pa1._Property.SetColor(1, 0.5, 0.5)
    pa1.bindDataset(pa)
    pa.visible(0)


    def f(**kwargs):
        pprint(kwargs)


    for b in pa.mouse:
        for e in pa.mouse[b]:
            if 'Move' not in e:
                pa.mouse[b][e] = f

    app.exec_()

    import shelve
    from pickle import HIGHEST_PROTOCOL

    with shelve.open('C:/Users/Medraw/Desktop/db', 'n', HIGHEST_PROTOCOL) as db:
        db['w'] = {'Vi': w, 'pa': pa, 'pa1': pa1}

    with shelve.open('C:/Users/Medraw/Desktop/db', 'r') as db:
        db = db['w']
        db['Vi'].show()
        sys.exit(app.exec_())
